/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#ifndef ___DATA_H
#define ___DATA_H

#include <vdr/tools.h>
#include <vdr/thread.h>

#define MBC(a,b) (((a)<<8)+(b))

class cLoaders;
class cPlainKeys;

// ----------------------------------------------------------------

class cLoader : public cListObject {
friend class cLoaders;
private:
  bool modified;
  char *id;
protected:
  void Modified(bool mod=true) { modified=mod; }
public:
  cLoader(const char *Id);
  virtual ~cLoader();
  virtual bool ParseLine(char *line, bool fromCache)=0;
  virtual bool Save(FILE *f)=0;
  bool IsModified(void) { return modified; }
  const char *Id(void) { return id; }
  };

// ----------------------------------------------------------------

class cLoaders : public cMutex, cList<cLoader> {
private:
  char *cacheFile;
  //
  cLoader *FindLoader(const char *id);
  bool IsModified(void);
public:
  cLoaders(void);
  ~cLoaders();
  void Register(cLoader *ld);
  void LoadCache(const char *cfgdir);
  void SaveCache(void);
  };

extern cLoaders loaders;

// ----------------------------------------------------------------

class cECMInfo : public cListObject {
public:
  char *name;
  int system;
  int ecm_pid;
  int id;
  //
  cECMInfo(void);
  cECMInfo(cECMInfo *e);
  cECMInfo(char *Name, int System, int Pid, int Id);
  ~cECMInfo();
  };

// ----------------------------------------------------------------

class cEcmData : public cECMInfo {
public:
  int prgId, source, transponder;
  //
  cEcmData(void);
  cEcmData(int PrgId, int Source, int Transponder, cECMInfo *e);
  bool Save(FILE *f);
  bool Parse(const char *buf);
  };

// ----------------------------------------------------------------

class cEcmCache : public cLoader, cMutex, cList<cEcmData> {
private:
  bool Exists(int PrgId, int Source, int Transponder, cECMInfo *e);
public:
  cEcmCache(void);
  void Load(void);
  void New(int PrgId, int Source, int Transponder, cECMInfo *e);
  cEcmData *Search(int PrgId, int Source, int Transponder, cEcmData *dat=0);
  virtual bool Save(FILE *f);
  virtual bool ParseLine(char *line, bool fromCache);
  };

// ----------------------------------------------------------------

class cPlainKey : public cListObject {
friend class cPlainKeys;
private:
  bool au, del;
protected:
  void SetInvalid(void) { del=true; }
  void SetAuto(void) { au=true; }
  bool IsAuto(void) { return au; }
  bool IsInvalid(void) { return del; }
  bool Set(int Type, int Id, int Keynr, void *Key);
  virtual bool SetKey(void *Key)=0;
public:
  int type, id, keynr;
  //
  cPlainKey(void);
  virtual bool Parse(char *line)=0;
  virtual bool Save(FILE *f)=0;
  virtual char *Print(void)=0;
  virtual bool Cmp(void *Key)=0;
  virtual void Get(void *mem)=0;
  };

// ----------------------------------------------------------------

class cPlainKeys : public cLoader, cMutex, cList<cPlainKey> {
private:
  cPlainKey *NewFromType(int type);
public:
  cPlainKeys(void);
  bool Load(const char *cfgdir);
  virtual bool Save(FILE *f);
  virtual bool ParseLine(char *line, bool Au=true);
  cPlainKey *FindKey(int Type, int Id, int Keynr, cPlainKey *key=0);
  bool NewKey(int Type, int Id, int Keynr, void *Key);
  };

#endif //___DATA_H
