/*
    iso.c
    ISO input/output functions
*/

#include <stdio.h>
#include <assert.h>
#include <memory.h>
#include "sio.h"
#include "iso.h"

#define ATR_TO  500
#define INS_TO  1000
#define DATA_TO 4000

#ifdef DEBUG
#define d(a)  {(a);}

static void printHEX(uchar val)
{
	Log ("%02X", val);
}

static void DumpHex(uchar *ptr, int len)
{
	while (len--) {
		printHEX(*ptr++);
		Log (" ");
	}
}


#else
#define d(a)
#endif


static int i_fd = -1;

int i_reset(uchar *atr)
{
  // buffer flushen, 
	// reset puls ausloesen 
	sio_cardreset(i_fd);

	d(Log(" <- ATR: "));
	if(sio_read(i_fd, ATR_TO, ATR_LEN, atr))
		return 1;
	d(DumpHex(atr, 20));
	d(Log("\n"));	
	
	return 0;
}


int i_iso_close()
{
	CloseLog() ;
  sio_close(i_fd);
  i_fd = -1;
}

int i_iso_open(char *pc_dev)
{
	if (sio_init(&i_fd, pc_dev))
		return 1;
	
	return 0;
}

static uchar CalcChecksum(const uchar *buffer, const int bufferlen)
{
	int		i;

	uchar ixor = 0x3f;
	for (i=0 ; i < bufferlen;i++) {
		ixor ^= *(buffer+i);	// bad performance - who cares
	}	
  return ixor;
}

// Es wird der String ohne Checksum uebergeben. 
static int i_iso_cmd (uchar *cmdstring, int cmdstringlen)
{
	char	cmdbuffer[129];
  uchar ixor;

	assert(cmdstringlen <= 128);

	d(Log(" -> INS: "));
	if (sio_write(i_fd, 0, cmdstringlen, cmdstring))
  	return -1;
  
  ixor = CalcChecksum(cmdstring, cmdstringlen);

	if (sio_write(i_fd, 0, 1, &ixor))
		return -1;

	/* Hier jetzt gesendete Daten zuruecklesen */
	if (sio_read(i_fd, DATA_TO, cmdstringlen+1, (char*)cmdbuffer))
		return -1;

  if ( memcmp (cmdbuffer, cmdstring, cmdstringlen)) {
    return -1;
  }

	d(DumpHex(cmdbuffer, cmdstringlen+1 ));
	d(Log("\n"));
	return 0;
}

// Return: wenn > 0, laenge des telegrams mit Checksum
int i_DoCmd(ISOCmd *cmd, uchar *response, int maxresponse)
{
  uchar checksum;
  int len;

  // Ausgabe des Kommandos
  if (!i_iso_cmd ((uchar*)cmd, cmd->len + 6)) {
		d(Log(" <- RES: "));
		if (sio_read(i_fd, DATA_TO, 9, response)) 
        return -1;  
		len = 9;
		if (response[7]) {
			if (sio_read(i_fd, DATA_TO, response[7], &response[9])) 
				return -1;  
			len += response[7];
		}
    checksum = CalcChecksum(response, 9 + response[7]);
		if (checksum) {
			d(Log("Checksum Error!\n"));
			return -1;
		}
  } else {
    return -1;
  }
  d(DumpHex(response, len));
  d(Log ("Len: %d\n",len));
  return len;
}

int i_card_inserted(void)
{
  return sio_card_inserted(i_fd);
}

