/*
    iso.c
    ISO input/output functions
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <fcntl.h>

#include <sys/ioctl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "sio.h"
#include "iso.h"

#define ATR_TO 500
#define INS_TO 500
#define DATA_TO 1000

#ifdef DEBUG
#define d(a) a
#else
#define d(a)
#endif

void v_dump(int i_len, uchar *pc_data);

static int i_fd = -1;

int i_iso_open(char *pc_dev, uchar *atr)
{
	int i;

	if (sio_init(&i_fd, pc_dev))
		return 1;

	d(printf(" <- ATR: "));
	if(sio_read(i_fd, ATR_TO, ATR_LEN, atr))
		return 1;

	return 0;
}

int i_iso_write(const uchar *pc_ins, uchar *pc_data, const uchar *pc_sw)
{
	int i_ret;
	int i = 0;
	uchar pc_ack[MAX_LEN + 1];

	d(printf(" -> INS: "));
	if (sio_write(i_fd, 0, INS_LEN, pc_ins))
		return 1;
	if (sio_read(i_fd, INS_TO, INS_LEN + 1, pc_ack))
		return 1;
	v_dump(INS_LEN + 1, pc_ack);
	/* XXX compare ack a ins */
	if (pc_ack[INS_LEN + 1] == INS_ACK(pc_ins))
		return 1;

	d(printf(" <- DATA: "));
	if (sio_write(i_fd, 0, DATA_LEN(pc_ins), pc_data))
		return 1;
	if (sio_read(i_fd, DATA_TO, DATA_LEN(pc_ins) + SV_LEN, pc_ack))
		return 1;
	v_dump(DATA_LEN(pc_ins) + SV_LEN, pc_ack);
	/* XXX compare ack a ins */
	if (pc_ack[DATA_LEN(pc_ins)] != pc_sw[0] ||
		pc_ack[DATA_LEN(pc_ins) + 1] != pc_sw[1])
		return 1;

	/* XXX why this?
        memcpy(pc_sw, pc_ack + DATA_LEN(pc_ins), SV_LEN);
        */

	return 0;
}

int i_iso_read(const uchar *pc_ins, uchar *pc_data, const uchar *pc_sw)
{
	uchar pc_ack[MAX_LEN + 1];

	d(printf(" -> INS: "));
	if (sio_write(i_fd, 0, INS_LEN, pc_ins))
		return 1;
	if (sio_read(i_fd, INS_TO, INS_LEN + 1, pc_ack))
		return 1;
	v_dump(INS_LEN + 1, pc_ack);
	/* XXX compare ack a ins */
	if (pc_ack[INS_LEN + 1] == INS_ACK(pc_ins))
		return 1;

	d(printf(" <- DATA: "));
	if (sio_read(i_fd, DATA_TO, DATA_LEN(pc_ins) + SV_LEN, pc_ack))
		return 1;
	v_dump(DATA_LEN(pc_ins) + SV_LEN, pc_ack);
	/* XXX compare ack a ins */
	if (pc_ack[DATA_LEN(pc_ins)] != pc_sw[0] ||
		pc_ack[DATA_LEN(pc_ins) + 1] != pc_sw[1])
		return 1;
	memcpy(pc_data, pc_ack, DATA_LEN(pc_ins));

	return 0;
}

void v_dump(int i_len, uchar *pc_data)
{
#ifdef DEBUG
	int i;

	for (i=0; i < i_len; i++)
		d(printf("%02X ", pc_data[i]));
	d(printf("\n"));
#endif
}
