#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/mman.h>
#include <sys/stat.h>

//#include "roms.h"
#include "nagra.h"

#define MAX_ROMS 11

struct Rom { unsigned char *addr;
             int len, fd;
             bool failed;
           };

static pthread_mutex_t rommutex=PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
static struct Rom Roms[MAX_ROMS] = {};
static char *RomDir=0;

static void LoadRom(int romNr)
{
  pthread_mutex_lock(&rommutex);
  struct Rom *r=&Roms[romNr-1];
  if(!r->addr && !r->failed) {
    char path[256];
    snprintf(path,sizeof(path),"%s/ROM%d.bin",RomDir,romNr);
    struct stat64 ds;
    if(!stat64(path,&ds)) {
      if(S_ISREG(ds.st_mode)) {
        int fd=open(path,O_RDONLY);
        if(fd>=0) {
          unsigned char *map=(unsigned char *)mmap(0,ds.st_size,PROT_READ,MAP_SHARED,fd,0);
          if(map!=MAP_FAILED) {
            r->addr=map; r->len=ds.st_size; r->fd=fd;
            pthread_mutex_unlock(&rommutex);
            return;
            }
          close(fd);
          }
        }
      }
    r->failed=true; // don't try this one over and over again
    printf("missing/failed ROM%d.bin\n",romNr);
    }
  pthread_mutex_unlock(&rommutex);
}

void FreeRoms(void)
{
  pthread_mutex_lock(&rommutex);
  for(int i=0 ; i<MAX_ROMS ; i++) {
    struct Rom *r=&Roms[i];
    if(r->addr) {
      munmap(r->addr,r->len);
      close(r->fd);
      r->addr=0;
      }
    }
  pthread_mutex_unlock(&rommutex);
}

void SetRomDir(const char *romdir)
{
  free(RomDir);
  RomDir=strdup(romdir);
}

static const unsigned char *GetRomBytes(int romNr, int offset, int count)
{
  unsigned char *loc=0;
  pthread_mutex_lock(&rommutex);
  struct Rom *r=&Roms[romNr-1];
  if(!r->addr) LoadRom(romNr);
  if(r->addr && offset>=0 && offset+count<r->len) {
    loc=r->addr+offset;
    }
  pthread_mutex_unlock(&rommutex);
  return loc;
}

int
Nagra_GetKeys(unsigned char *_decodedEMM, unsigned char *_provider, unsigned char *_key0, unsigned char *_key1)
{
    // returns:
    // 0 - Not recognised as a key update EMM
    // 1 - Keys updated successfully
    // 2 - missing ROM

    unsigned char hash[16];
    unsigned char block[64];

    int tester=0;
    for(int i=0; i<4; i++) tester = (tester<<8) + _decodedEMM[i+3];

    int romNr;
    switch (tester) {
      case 0xf3cd6a0a: romNr=3; break;
      case 0xf4cd6909: romNr=4; break;
      case 0xf7cd584d: romNr=7; break;
      case 0xfacd7ab7: romNr=10; break;
      case 0xfbcd456f: romNr=11; break;
      default: return 2;
      }
    
    const int offset = ((_decodedEMM[7]<<8)+_decodedEMM[8]) - 0x4000;
    const unsigned char *addr=GetRomBytes(romNr,offset,32);
    if(!addr) return 2;
    memcpy(block,addr,32);

    memcpy(hash,block,8);
    Nagra_Hash(block, hash, 4);

    if ((_decodedEMM[44] == 0x42) && (_decodedEMM[45] == 0x05)) {

	memcpy(_provider, _decodedEMM + 42, 2);	//the ident for which the new keys are intended

	memcpy(block, _decodedEMM + 46, 8);	// use the hash & E*DES to arrive at the new plainkeys
	Nagra_KeyCrypt(block, hash, _key0);

	memcpy(block, _decodedEMM + 56, 8);
	Nagra_KeyCrypt(block, hash, _key1);

	return 1;
    }

    if ((_decodedEMM[36] == 0xCC) && (_decodedEMM[37] == 0x5B)
	&& (_decodedEMM[38] == 0x27)) {
	unsigned char *pBlock = _decodedEMM + 39;	// DISH / BEV style

	for (int i = 0; i < 3; i++)
	    Nagra_KeyCrypt(pBlock + i * 8, hash, block + i * 8);

	memcpy(_provider, block + 2, 2); // ident for which  the new keys are intended
	memcpy(_key0, block + 6, 8);	 // use the hash and E*DES to arrive at the new plainkeys
	memcpy(_key1, block + 16, 8);

	return 1;
    }
    return 0;
}

