
#include <stdlib.h>
#include <string.h>
#include "nagra.h"

bool
Nagra_RSADecrypt_ECM(unsigned char *data, int keyselect, int pmecmlen,
		     BIGNUM * e1, BIGNUM * n1, BIGNUM * n2)
{
    BN_CTX *ctx = BN_CTX_new();
    BIGNUM *e2 = BN_new();
    BIGNUM *result = BN_new();
    int ks = keyselect;

    BN_hex2bn(&e2, "03");
    RotateBytes(data, pmecmlen);
    BIGNUM *ecm = BN_bin2bn(data, pmecmlen, NULL);

    bool res=true;

    if (!BN_mod_exp(result, ecm, e2, n2, ctx)) {
	printf("Error in round 1 of RSA Decrypt\n");
        res=false; goto out;
    }

    ks = ks >> 4;
    result->d[result->top - 1] |= ks << 28;

    if (!BN_mod_exp(result, result, e1, n1, ctx)) {
	printf("Error in round 2 of RSA Decrypt\n");
        res=false; goto out;
    }

    if (BN_bn2bin(result, data) < 64) {
	printf("There was an unexpected error.\n");
        res=false; goto out;
    }

out:
    if (result) BN_free(result);
    if (e2) BN_free(e2);
    if (ecm) BN_free(ecm);
    if (ctx) BN_CTX_free(ctx);
    return res;
}

int
Nagra_RSADecrypt_EMM(unsigned char *data, unsigned char *signature,
		     int keyselect, unsigned char *vkey, int emmlen,
		     BIGNUM * e1, BIGNUM * n1, BIGNUM * n2)
{

    BN_CTX *ctx = BN_CTX_new();
    BIGNUM *e2 = BN_new();
    BIGNUM *result1 = BN_new();
    BIGNUM *result2 = BN_new();

    RotateBytes(data, emmlen);	// LE <-> BE
    RotateBytes(signature, 8);

    BIGNUM *emm = BN_bin2bn(data, emmlen, NULL);
    BIGNUM *sig = BN_bin2bn(signature, 8, NULL);

    BN_hex2bn(&e2, "03");

    int res=0;
    int ks = keyselect;
    unsigned long newsigH, newsigL;	// used for hash test code

    if (!BN_mod_exp(result1, emm, e2, n2, ctx)) {
	printf("Error in round 1 of RSA Decrypt EMM\n");
        res=0; goto out;
    }

    ks = ks >> 6;
    result1->d[result1->top - 1] |= ks << 30;

    if (!BN_mod_exp(result2, result1, e1, n1, ctx)) {
	printf("Error in round 2 of RSA Decrypt EMM\n");
        res=0; goto out;
    }

    if (BN_bn2bin(result2, data) < 64) {
        res=0; goto out;
    }

    if (Nagra_SigCheck(data, signature, vkey, 8)) {
    	res=1; goto out;
    }

    // We might need to use intercambia firma method...
    newsigH = result1->d[0x01];	// Intercambia firma:
    newsigL = result1->d[0x00];	// Exchange signature for lower 64 bits after round 1 powmod()

    for (int i = 0; i < 4; i++) {	// build revised signature in reverse
	signature[i] = ((0xff << (i * 8)) & newsigL) >> (i * 8);
    }				// byte order of result1 lower 64 bits
    for (int i = 0; i < 4; i++) {
	signature[4 + i] = ((0xff << (i * 8)) & newsigH) >> (i * 8);
    }

    result1->d[0x01] = sig->d[0x01];	// use reverse order sig as N2 lower 64 bits
    result1->d[0x00] = sig->d[0x00];	// this is the "intercambia firma" modification

    if (!BN_mod_exp(result2, result1, e1, n1, ctx)) {
	printf("Error in round 3 of RSA Decrypt EMM\n");
        res=0; goto out;
    }

    if (BN_bn2bin(result2, data) < 64) {
        res=0; goto out;
    }

    if (Nagra_SigCheck(data, signature, vkey, 8)) {
        res=2; goto out;
    }

out:
    if (result1) BN_free(result1);
    if (result2) BN_free(result2);
    if (e2) BN_free(e2);
    if (emm) BN_free(emm);
    if (ctx) BN_CTX_free(ctx);
    return res;
}

