/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#ifdef NAGRA_SUPPORT

#include <stdlib.h>
#include <stdio.h>
#include <openssl/bn.h>

#include "common.h"
#include "logger-nagra.h"
#include "logger.h"
#include "softcam.h"
#include "data.h"
#include "libnagra/nagra.h"

// -- cLogNagra ------------------------------------------------------------------

cLogNagra::cLogNagra(int CardNum)
:cLogSys(CardNum)
{}

int cLogNagra::ParseCAT(unsigned char *buffer, int n, int sys, int id)
{
  for(int index=8 ; index<n ;) {
    if(buffer[index]==0x09) {
      int s=WORD(buffer,index+2,0xFFFF);
      if(sys==s) return WORD(buffer,index+4,0x1FFF);
      }
    index+=buffer[index+1]+2;
    }
  return -1;

}
int cLogNagra::Section(void)
{
  return 0x82;
}

void cLogNagra::Log(unsigned char *buffer, int n)
{
  unsigned char v[8];
  int id=buffer[10]*256 + buffer[11];
  cPlainKey *pk;

  // no FastKey() here!
  BIGNUM *e1 = BN_new();
  if(!(pk=keys->FindKey('N',id,MBC('E','1')))) return;
  pk->Get(e1);
  BIGNUM *n1 = BN_new();
  if(!(pk=keys->FindKey('N',id,MBC('N','1')))) return;
  pk->Get(n1);
  BIGNUM *n2 = BN_new();
  if(!(pk=keys->FindKey('N',id,MBC('N','2')))) return;
  pk->Get(n2);

  if(!(pk=keys->FindKey('N',id,'V'))) return;
  pk->Get(v);
  
  if(Nagra_RSADecrypt_EMM(&buffer[21],&buffer[13],(int)buffer[12],v,64,e1,n1,n2)) {
    unsigned char key0[8], key1[8], ident[2];
    switch(Nagra_GetKeys(buffer+21,ident,key0,key1)) {
      case 1: // keys found
        dl(printf("logger %d: got Nagra keys for %04X\n",cardNum,id))
        cSoftCAM::FoundKey();
        if(keys->NewKey('N',id,00,key0)) cSoftCAM::NewKey();
        cSoftCAM::FoundKey();
        if(keys->NewKey('N',id,01,key1)) cSoftCAM::NewKey();
        cSoftCAM::Save();
        break;
      case 2: // rom file missing
        dl(printf("logger %d: missing ROM file\n",cardNum))
        // fall through
      case 0: // no valid key found
        break; 
      default:
        dl(printf("logger %d: unexpected error\n",cardNum))
        break;
      }
    } 
  BN_free(e1); BN_free(n1); BN_free(n2);
}

#endif //NAGRA_SUPPORT
