/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

#include "common.h"
#include "logger-seca.h"
#include "logger.h"
#include "softcam.h"
#include "data.h"
#include "misc.h"

struct Seca {
  unsigned char sep0, sep1;
  unsigned char cmdLen;
  unsigned char id[2];
  unsigned char sa[4];
  unsigned char keyNr;
  unsigned char data[78];
  };

// -- cSecaCardInfo ----------------------------------------------------------------

class cSecaCardInfo : public cListObject {
public:
  unsigned char ID[2], SA[4], KEY[8];
  //
  bool Parse(char *line);
  bool Save(FILE *f) { return true; }
  bool IsUpdated(void) { return false; }
  void Updated(void) {}
  bool Cmp(cSecaCardInfo *ci) { return false; }
  };

bool cSecaCardInfo::Parse(char *line)
{
  while(*line && isspace(*line)) line++;
  if(*line!=';') {
    if(GetHex(line,ID,sizeof(ID)) && GetHex(line,SA,sizeof(SA)) && GetHex(line,KEY,sizeof(KEY)))
      return true;
    }
  return false;
}

// -- cSecaCardInfos ---------------------------------------------------------------

class cSecaCardInfos : public cCardInfos<cSecaCardInfo> {
public:
  cSecaCardInfos(void);
  };

cSecaCardInfos::cSecaCardInfos(void)
:cCardInfos<cSecaCardInfo>("Seca")
{}

cSecaCardInfos *Scards=0;

// -- cLogSeca ------------------------------------------------------------------

cLogSeca::cLogSeca(int CardNum)
:cLogSys(CardNum)
{
}

bool cLogSeca::HaveCards(void)
{
  return Scards->HaveCards();
}

bool cLogSeca::LoadCards(const char *cfgdir)
{
  if(!Scards) Scards=new cSecaCardInfos;
  if(!Scards) {
    esyslog("ERROR: no memory for Seca cards!");
    return false;
    }
  Scards->Load(cfgdir,"Seca","Seca.KID");
  return true;
}

int cLogSeca::ParseCAT(unsigned char *buffer, int n, int sys, int id)
{
  for(int index=8 ; index<n ;) {
    if(buffer[index]==0x09) {
      if(WORD(buffer,index+2,0xFFFF)==sys && buffer[index+1]>4) {
        int i=index+7;
        for(int nn=buffer[index+6] ; nn ; nn--) {
          if(WORD(buffer,i+2,0xFFFF)==id) {
            currId=id;
            return WORD(buffer,i,0x1FFF);
            }
          i+=4;
          }
        }
      }
    index+=buffer[index+1]+2;
    }
  return -1;
}

int cLogSeca::Section(void)
{
  return 0x84;
}

void cLogSeca::Log(unsigned char *buffer, int n)
{
  struct Seca *s=(struct Seca *)buffer;
  while(n>=(int)sizeof(struct Seca)) {
    if(s->cmdLen==0x55) {
      cSecaCardInfo *ci=Scards->First();
#ifdef DEBUG_LOG
      bool first=true;
#endif
      while(ci) {
        if(((s->id[0]==ci->ID[0] && s->id[1]==ci->ID[1]) || (!ci->ID[0] && !ci->ID[1])) &&
           ((s->sa[0]==ci->SA[0] && s->sa[1]==ci->SA[1] && s->sa[2]==ci->SA[2]) || (!ci->SA[0] && !ci->SA[1] && !ci->SA[2] && !ci->SA[3])) ) {
#ifdef DEBUG_LOG
          char str[20],str2[20],str3[20];
          if(first) {
            printf("\rID %s SA %s",HexStr(str,s->id,2),HexStr(str2,s->sa,4));
            first=false;
            }
#endif

          unsigned char MK[16];
          memcpy(&MK[0],ci->KEY,8);
          memcpy(&MK[8],ci->KEY,8);
          unsigned char buff[78];
          memcpy(&buff[0],&s->data[0],sizeof(buff));
          for(int i=0 ; i<=64 ; i+=8) decrypt_seca(MK,&buff[i]);

          unsigned char signature[8];
          unsigned char keyN[4], *key[4];
          unsigned int numKeys=0;
          bool sigOk=false;
          int pos, len;
          for(pos=0 ; pos<78 ; pos+=len) {
            int cmd=buff[pos++];
            len=(cmd>>4)&0x0f;
            switch(len) {
              case 0x0b: len=0x0e; break;
              case 0x0d: len=0x10; break;
              case 0x0e: len=0x18; break;
              case 0x0f: len=0x20; break;
              }
            switch(cmd) {
              case 0x82: // signature
                memset(signature,0,sizeof(signature));
                for(unsigned int i=0 ; i<sizeof(buff)-8 ; i+=8) {
                  for(unsigned int j=0 ; j<8 && i+j<sizeof(buff)-8; j++) signature[j]^=buff[i+j];
                  encrypt_seca(MK,signature);
                  }
                if(!memcmp(signature,&buff[pos],sizeof(signature))) sigOk=true;
                else pos=999;
                break;

              case 0x90: // new primary key
                if(numKeys<sizeof(keyN)) {
                  keyN[numKeys]=buff[pos] & 0x0f;
                  key[numKeys]=&buff[pos+1];
                  numKeys++;
                  }
                break;
              }
            }
          if(pos==78 && sigOk) {
            dl(printf(" - OK (CI ID %s SA %s KEY01 %s)",HexStr(str,ci->ID,2),HexStr(str2,ci->SA,4),HexStr(str3,ci->KEY,8)))
#ifdef EMM_DUMP
            Dump(EMM_DUMP,buffer,n);
#endif
#ifdef DEBUG_LOG_FILE
            FILE *f=LogStart();
            if(f) {
              char str[64],str2[20],str3[20];
              fprintf(f,"ID %s SA %s -  OK ",HexStr(str,s->id,2),HexStr(str2,s->sa,4));
              fprintf(f," with CI ID %s SA %s KEY01 %s\n",HexStr(str,ci->ID,2),HexStr(str2,ci->SA,4),HexStr(str3,ci->KEY,8));
              fprintf(f,"     cmds:");
              for(pos=0 ; pos<78 ; pos+=len) {
                int cmd=buff[pos++];
                len=(cmd>>4)&0x0f;
                switch(len) {
                  case 0x0b: len=0x0e; break;
                  case 0x0d: len=0x10; break;
                  case 0x0e: len=0x18; break;
                  case 0x0f: len=0x20; break;
                  }
                fprintf(f," 0x%02x/%d",cmd,len);
                }
              fprintf(f,"\n");
              for(int i=0 ; i<80 ; i+=16)
                fprintf(f,"     %02x %s\n",i,HexStr(str,&buff[i],i+16>78?78-i:16));
              LogEnd(f);
              }
#endif

            for(unsigned int i=0 ; i<numKeys ; i++) {
              decrypt_seca(MK,key[i]);
              dl(printf(" KEY %02x %s",keyN[i],HexStr(str,key[i],8)))
              cSoftCAM::FoundKey();
              if(keys->NewKey('S',currId,keyN[i],key[i])) {
                cSoftCAM::NewKey();
#ifdef DEBUG_LOG_FILE
                FILE *f=LogStart();
                if(f) {
                  char str[20];
                  fprintf(f,"New key %02x %s\n",keyN[i],HexStr(str,key[i],8));
                  LogEnd(f);
                  }
#endif
                }
              }
            cSoftCAM::Save();

            dl(printf("\n"))
            break;
            }
          else if((ci->ID[0]+ci->ID[1])!=0 && (ci->SA[0]+ci->SA[1]+ci->SA[2]+ci->SA[3])!=0) {
#ifdef DEBUG_LOG
            printf(" - FAIL\n");
            first=true;
#endif
#ifdef DEBUG_LOG_FILE
            FILE *f=LogStart();
            if(f) {
              char str[20],str2[20],str3[20];
              fprintf(f,"ID %s SA %s - FAIL",HexStr(str,s->id,2),HexStr(str2,s->sa,4));
              fprintf(f," with CI ID %s SA %s KEY01 %s\n",HexStr(str,ci->ID,2),HexStr(str2,ci->SA,4),HexStr(str3,ci->KEY,8));
              LogEnd(f);
              }
#endif
            }
          dl(fflush(stdout))
          }
        ci=Scards->cList<cSecaCardInfo>::Next(ci);
        }
      }
    else {
#ifdef EMM_DUMP
      Dump(EMM_DUMP"_bad",buffer,n);
#endif
      }
    n-=sizeof(struct Seca);
    s++;
    }
}
