/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#ifndef ___LOGGER_H
#define ___LOGGER_H

#include <vdr/thread.h>
#include <vdr/tools.h>
#include "data.h"
#include "misc.h"

// ----------------------------------------------------------------

class cLogSys {
protected:
  int cardNum;
public:
  cLogSys(int CardNum) { cardNum=CardNum; }
  virtual void Log(unsigned char *buffer, int n)=0;
  virtual bool HaveCards(void)=0;
  virtual int ParseCAT(unsigned char *buffer, int n, int sys, int id)=0;
  virtual int Section(void)=0;
  };

// ----------------------------------------------------------------

class cLogger : public cThread, cRunMode {
private:
  int emmDev;
  int caSys, newCaSys, Id, newId, emmPid;
  int cardNum;
  cLogSys *currentLogger;
  //
  virtual void Action(void);
  int ParseCAT(void);
public:
  cLogger(int CardNum, int EmmDev);
  ~cLogger();  
  void SetCaSys(int CaSys, int ID);
  static bool LoadCards(const char *cfgdir);
  };

// ----------------------------------------------------------------

template<class T> class cCardInfos : public cList<T>, public cLoader {
private:
  const char *sysname;
public:
  cCardInfos(const char *id):cLoader(id) {}
  bool HaveCards(void) { return (Count()!=0); }
  bool Load(const char *cfgdir, const char *SysName, const char *kidName)
    {
    sysname=SysName;
    Clear();
    const char *name=AddDirectory(cfgdir,kidName);
    FILE *f=fopen(name,"r");
    if(f) {
      isyslog("loading %s card infos from %s",SysName,name);
      char buff[512];
      while(fgets(buff,sizeof(buff),f)) {
        if(buff[0]=='*') break;    // ignore rest of file
        if(buff[0]==';') continue; // skip comment lines
        if(!ParseLine(buff,false)) break;
        }
      fclose(f);
      }
    d(printf("cardinfo: loaded %d %s cards from %s\n",Count(),SysName,name))
    return HaveCards();
    }
  virtual bool ParseLine(char *line, bool fromCache)
    {
      T *k=new T;
      if(k) {
        if(k->Parse((char *)line)) {
          if(!fromCache) { Add(k); k=0; }
          else {
            T *o=First();
            while(o) {
              if(o->Cmp(k)) {
                //d(printf("cardinfo: replacing from cache\n"))
                Add(k,o); k->Updated(); k=0;
                Del(o);
                break;
                }
              o=cList<T>::Next(o);
              }
            }
          }
        delete k;
        return true;
        }
      esyslog("ERROR: no enough memory for %s card infos!",sysname);
      return false;
    }
  virtual bool Save(FILE *f)
    {
    bool res=true;
    T *k=First();
    while(k) {
      if(k->IsUpdated() && !k->Save(f)) { res=false; break; }
      k=cList<T>::Next(k);
      }
    Modified(!res);
    return res;
    }
  };

#endif //___LOGGER_H
