/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#ifndef ___SOFTCAM_H
#define ___SOFTCAM_H

#include <vdr/thread.h>
#include "data.h"
#include "misc.h"

#define MAXDVBAPI 4

#define MASK_CA_SYSTEM      0xFF00

#define SECA_CA_SYSTEM      0x0100
#define VIACCESS_CA_SYSTEM  0x0500
#define IRDETO_CA_SYSTEM    0x0600
#define BETA_CA_SYSTEM      0x1700 
#define NAGRA_CA_SYSTEM     0x1800
#define NAGRA_BEV_CA_SYSTEM 0x1200
#define CRYPTOW_CA_SYSTEM   0x0D00
#define CONAX_CA_SYSTEM     0x0B00

#define I_CHID_SIZE 10

// ----------------------------------------------------------------

extern cEcmCache *ecmcache;
extern cPlainKeys *keys;

extern "C" {
  void decrypt_seca(unsigned char *k,unsigned char *d);
  void encrypt_seca(unsigned char *k,unsigned char *d);
  void sessionKeyCrypt(unsigned char *data,unsigned char *key,int date);
  int signatureCheck(unsigned char *data,int length,unsigned char *key,int date,unsigned char *signature,int keylen);
  void decryptIrd(unsigned char *data,unsigned char *key,int rounds,int offset);
  void rotate(unsigned char *p);
  int decrypt(unsigned char work_key[], unsigned char rx_message[], int rx_message_length, unsigned char des_data1[], unsigned char des_data2[]);
  int cryptoworks_decode(unsigned char *cws, unsigned char *data, const char *portname);
  }

// ----------------------------------------------------------------

class cDevice;

class cECMInfo;
class cECM;
class cLogger;
class cPlainKey;
class cSoftCAM;

// ----------------------------------------------------------------

struct CAMInfo {
  cSoftCAM *cam;
  int pid;
  bool shutUp;
  };

class cSoftCAM : public cThread, cRunMode {
public:
  static struct CAMInfo camInfo[MAXDVBAPI];
  static cMutex camLock;
private:
#ifdef NAGRA_SUPPORT
  unsigned char Nagra_MECMTable[256];
#endif
#ifdef CRYPTOWORKS_SUPPORT
  static cMutex cryptow_lock;
#endif
  bool i_flag;
  int i_chid[I_CHID_SIZE];
  //
  int firstKey;
  cPlainKey *currentKey;
  int lastType, lastId, lastKeynr;
  char currentKeyStr[48];
  unsigned char *cw, cwMem[16];
  //
  static int foundKeys, newKeys;
  static const char *cacheFile;
  //
  int cardNum;
  cDevice *device;
  int camDev, ecmDev, emmDev;
  bool softCSA;
  //
  cECM *ecm;
  cLogger *logger;
  int loggerTime;
  int prg, source, transponder, newprg, newSource, newTransponder;
  bool sync;
  //
  unsigned char lastEvenCW[8], lastOddCW[8];
  int dolog;
  //
  virtual void Action(void);
  void Init(void);
  void DeInit(void);
  bool Process(cECMInfo *ecm, unsigned char *buffer);
  bool PreCheck(cECMInfo *ecm, unsigned char *buffer);
  bool Irdeto(unsigned char *source);
  bool Seca(unsigned char *source, int length, int id, int keynr);
  bool Viaccess(unsigned char *source, int length, int id);
#ifdef CONAX_SUPPORT
  bool Conax(unsigned char *source, int length, int keynr);
  void ConaxParseECM(unsigned char *buf, int len);
#endif
#ifdef NAGRA_SUPPORT
  bool Nagra(unsigned char *data, int len);
  void ClearNagra(void);
  void WriteNagraTable(unsigned char *from, int off);
#endif
  void WriteCW(int index, bool resend);
  void SetLogger(int caSys, int Id, bool force=false);
  void KeyOK(cPlainKey *pk);
  void KeyFail(int type, int id, int keynr);
  void NoKey(void);
public:
  cSoftCAM(const cDevice *dev, int CardNum, int CamDev, int EcmDev, int EmmDev);
  ~cSoftCAM();
  void SetChannel(int Prg, int Source=0, int Transponder=0);
  cPlainKey *FastKey(int type, int id, int keynr);
  int CurrentPrg(void) { return prg; }
  //
  static cSoftCAM *Init(const cDevice *dev);
  static bool Startup(const char *cfgdir);
  static void ShutDown(void);
  static cSoftCAM *FindCAM(const cDevice *dev);
  static void Save(void);
  //
  static void FoundKey(void) { foundKeys++; }
  static void NewKey(void) { newKeys++; }
  static void KeyStats(int &fk, int &nk) { fk=foundKeys; nk=newKeys; }
  static const char *CurrKeyStr(int CardNum);
  static bool Active(void);
  };

#endif // ___SOFTCAM_H
